# dependencies
import nanoid
import json
import bson
import os
import random_utilities

from random_utilities.models.time_created import TimeCreatedModel

class DataModel:
    def __init__(self):
        self.key = nanoid.generate()
        self.time_created = TimeCreatedModel().__dict__
        self.last_modified = TimeCreatedModel().__dict__
        self._selected_database_ = "live" if os.environ.get("ENVIRONMENT") == "production" else "testing"

    def to_json(self) -> str:            
        return json.dumps(obj=self.__dict__)

    def to_dict(self) -> dict:
        new_self_dict = {**self.__dict__}
        return self._obj_to_dict(new_self_dict)

    @classmethod
    def _obj_to_dict(cls, obj):
        random_utilities.log(f"Passed object type {type(obj)}")
        if type(obj) == dict:
            for parameter in obj:
                if type(obj[parameter]) == dict:
                    random_utilities.log(f"Re-sending back to converter {parameter} {obj}")
                    obj[parameter] = cls._obj_to_dict(obj[parameter])
                else:
                    allowed_objects = [str, bool, int, float, type(None)]
                    if type(obj[parameter]) == list:
                        random_utilities.log(f"Found a list {parameter}")
                        obj[parameter] = cls._obj_to_dict(obj[parameter])
                    elif type(obj[parameter]) not in allowed_objects:
                        random_utilities.log(f"Allowed, {type(obj[parameter])}")
                        obj[parameter] = str(obj[parameter])
        elif type(obj) == list:
            for index, _ in enumerate(obj):
                if type(obj[index]) == bson.objectid.ObjectId:
                    obj[index] = str(obj[index])
                else:
                    obj[index] = cls._obj_to_dict(obj[index])
        return obj

    def get_schema():
        return { }

    @classmethod
    def verify_schema(cls, d: dict, schema = None) -> list:
        schema = schema if schema is not None else cls.get_schema()
        schema_keys = schema.keys()
        errors = []
        for schema_key in schema_keys:
            if d.get(schema_key):
                if type(d.get(schema_key)) != schema[schema_key]:
                    errors.append({ "error": f'Invalid data type "{type(d.get(schema_key)).__name__}" used in {schema_key}. "{schema[schema_key].__name__}" required instead.', "error_type": "invalid"})
            else:
                errors.append({ "error": f"Attribute {schema_key} of {schema[schema_key].__name__} required in request body.", "error_type": "undefined" })
        return errors


http_codes = {
    # Informational.
    100: ('continue',),
    101: ('switching_protocols',),
    102: ('processing',),
    103: ('checkpoint',),
    122: ('uri_too_long', 'request_uri_too_long'),
    200: ('ok', 'okay', 'all_ok', 'all_okay', 'all_good', '\\o/', '✓'),
    201: ('created',),
    202: ('accepted',),
    203: ('non_authoritative_info', 'non_authoritative_information'),
    204: ('no_content',),
    205: ('reset_content', 'reset'),
    206: ('partial_content', 'partial'),
    207: ('multi_status', 'multiple_status', 'multi_stati', 'multiple_stati'),
    208: ('already_reported',),
    226: ('im_used',),

    # Redirection.
    300: ('multiple_choices',),
    301: ('moved_permanently', 'moved', '\\o-'),
    302: ('found',),
    303: ('see_other', 'other'),
    304: ('not_modified',),
    305: ('use_proxy',),
    306: ('switch_proxy',),
    307: ('temporary_redirect', 'temporary_moved', 'temporary'),
    308: ('permanent_redirect',
          'resume_incomplete', 'resume',),  # These 2 to be removed in 3.0

    # Client Error.
    400: ('bad_request', 'bad'),
    401: ('unauthorized',),
    402: ('payment_required', 'payment'),
    403: ('forbidden',),
    404: ('not_found', '-o-'),
    405: ('method_not_allowed', 'not_allowed'),
    406: ('not_acceptable',),
    407: ('proxy_authentication_required', 'proxy_auth', 'proxy_authentication'),
    408: ('request_timeout', 'timeout'),
    409: ('conflict',),
    410: ('gone',),
    411: ('length_required',),
    412: ('precondition_failed', 'precondition'),
    413: ('request_entity_too_large',),
    414: ('request_uri_too_large',),
    415: ('unsupported_media_type', 'unsupported_media', 'media_type'),
    416: ('requested_range_not_satisfiable', 'requested_range', 'range_not_satisfiable'),
    417: ('expectation_failed',),
    418: ('im_a_teapot', 'teapot', 'i_am_a_teapot'),
    421: ('misdirected_request',),
    422: ('unprocessable_entity', 'unprocessable'),
    423: ('locked',),
    424: ('failed_dependency', 'dependency'),
    425: ('unordered_collection', 'unordered'),
    426: ('upgrade_required', 'upgrade'),
    428: ('precondition_required', 'precondition'),
    429: ('too_many_requests', 'too_many'),
    431: ('header_fields_too_large', 'fields_too_large'),
    444: ('no_response', 'none'),
    449: ('retry_with', 'retry'),
    450: ('blocked_by_windows_parental_controls', 'parental_controls'),
    451: ('unavailable_for_legal_reasons', 'legal_reasons'),
    499: ('client_closed_request',),

    # Server Error.
    500: ('internal_server_error', 'server_error', '/o\\', '✗'),
    501: ('not_implemented',),
    502: ('bad_gateway',),
    503: ('service_unavailable', 'unavailable'),
    504: ('gateway_timeout',),
    505: ('http_version_not_supported', 'http_version'),
    506: ('variant_also_negotiates',),
    507: ('insufficient_storage',),
    509: ('bandwidth_limit_exceeded', 'bandwidth'),
    510: ('not_extended',),
    511: ('network_authentication_required', 'network_auth', 'network_authentication'),
}



# dependencies
import datetime
import months

class TimeCreatedModel():
    def __init__(self, timestamp=None):
        date = datetime.datetime.now()
        if timestamp:
            date = datetime.datetime.fromtimestamp(timestamp)
        month = months.Month(month=date.month, year=date.year)
        
        self.day = date.strftime('%d')
        self.month = month.month_abbr
        self.year = date.strftime('%Y')
        self.formatted_date = f'{self.day} {self.month}, {self.year}'
        self.time = date.strftime('%H:%M:%S')

        if timestamp:
            self.timestamp = timestamp
        else:
            self.timestamp = date.timestamp()



import os
import random
import pymongo
import console
from .models.time_created import TimeCreatedModel
import traceback
import sys


def log(msg: str, is_error=False, is_success=False, is_force=False):
    """ Logs verbose information about runtime. """
    if bool(os.environ.get("SHOW_LOGS", False)) == True or is_force:
        msg = console.fx.bold(str(msg))

        """Get the current timestamp"""
        _time = TimeCreatedModel()
        time = f"{_time.formatted_date} {_time.time}"

        """Show the message"""
        print(f'{console.fx.dim(time + " |")} [{console.fg.red(msg) if is_error else (console.fg.white(msg) if not is_success else console.fg.green(msg))}]')
log(f"""Verbose mode: {os.environ.get('SHOW_LOGS', 'False')}. To be able to use the log utility, requires you to set the 'SHOW_LOGS' env variable to 'True' / '1'.""", is_force=True)


def query_to_dict(query_string: str) -> dict:
    """ Convert a URL query string to a dictionary """
    query_string_split_pairs = query_string.split("&")
    dict_query_string = { }
    for query in query_string_split_pairs:
        query_split = query.split("=") # name=value === ["name", "value"]
        dict_query_string[query_split[0]] = query_split[1]
        if query_split[1] in ("true", "false"):
            dict_query_string[query_split[0]] = bool(query_split[1]) # Parse boolean values
        elif query_split[1].isdecimal():
            dict_query_string[query_split[0]] = float(query_split[1])
    return dict_query_string


def random_sort(_list: list) -> list:
    """ Returns a randomly sorted list. """
    random_sorted_indices = []
    for _ in _list:
        b = len(_list) - 1
        index = random.randint(a=0, b=b)
        while index in random_sorted_indices:
            index = random.randint(a=0, b=b)
            if index not in random_sorted_indices:
                break
        random_sorted_indices.append(index)
    
    """ Re-structure the list with the randomness. """
    random_list = [0 for i in range(len(_list))]
    for index, random_index in enumerate(random_sorted_indices):
        random_list[index] = _list[random_index]
    return random_list
    

def read_request_body(request) -> dict:
    """Reads the binary data sent in the request."""
    request_data = request.get_data()

    """Proof check if the data is readable as normal text or not."""
    if request_data != b'' or request_data.isascii():
        request_body = request.get_json()
        if request_body != None:
            return request_body
        else:
            """Else return an empty dictionary."""
            return dict()
    else:
        """Else return an empty dictionary."""
        return dict()


"""Checks if a dictionary contains all fields in a provided array, returns missing fields if any"""
def fields_all_check(dct: dict, lst: list) -> tuple:
    for field in dct:
        if field in lst:
            lst.remove(field)
    is_all_check = True if len(lst) == 0 else False
    return is_all_check, lst


"""Database connection capabilities."""
default_database = None # global access variable to save the db connections
default_collection = None # to assign it to a collection connection
def initiate_mongodb_connection(mongo_host, database_name, collection_name):
    global default_collection
    global default_database
    if mongo_host and collection_name and database_name:
        try:
            log("Connecting to database and testing connection...")
            mongo_client = pymongo.MongoClient(mongo_host,
                server_api=pymongo.server_api.ServerApi("1"))

            default_database = mongo_client[database_name]
            default_collection  = default_database[collection_name]
            document_count = default_collection.count_documents({ })

            log(f"Database successfully connected with `{document_count}` documents in collection `{collection_name}`.")

            return default_collection, True
        except:
            log("Opps something went wrong. " + traceback.format_exc())
            sys.exit(1)
    else:
        log("No database connection specified. Not connecting to any.")
        return None, False



import json
import socket
from flask import make_response
from random_utilities.models.http_codes import http_codes

class ResponseModel():
    def __init__(self, cd: int, rs=None, d=None, msg=None):
        try:
            # parse the code to a proper type
            self.status_code = str(cd)

            if msg != None:
                self.status_message = msg
            else:
                # propely format the response message
                self.status_message = ' '.join(
                    http_codes[cd][0].capitalize().split('_'))

            if rs != None:
                self.reason = rs

            if d != None:
                self.data = d

            # Incase one of the Pods is faulty in production
            self.pod = socket.gethostname()

        except KeyError as error:
            print('Invalid status code')

    def to_json(self):
        response = make_response(json.dumps(
            self.__dict__), int(self.status_code))
        response.headers['Content-Type'] = 'application/json'
        return response



from setuptools import find_packages, setup

setup(
    name="random_utilities",
    version="1.0.8",
    author="Hetchfund.Capital (Libby Lebyane)",
    author_email="<lebyane.lm@gmail.com>",
    description="Helper library for random operations",
    # package_dir={"": "random_utilities"},
    packages=["random_utilities", "random_utilities.models"],
    install_requires=["console", "months", "pymongo", "pymongo[srv]"],
    keywords=["random", "utilities", "tools", "helpers"]
)
